from z3 import *

# bounded model checking applied to Fischer's mutual exclusion protocol
# other examples can be checked by redefining the bad states, initial states, and step and time transition relations appropriately, etc.

# control locations defined as an enumeration datatype
ControlLocation = Datatype('Location')

ControlLocation.declare('idle')
ControlLocation.declare('start')
ControlLocation.declare('check')
ControlLocation.declare('cs')
ControlLocation = ControlLocation.create()

length = 1000 # number of loops to unroll

s = Solver() # instantiate a solver

opt_debug = False

x=[] # list of steps for continuous variable x
q=[] # list of steps for control location
g=[] # list of steps for global variable g

r = Int("r")
i, j, N = Ints("i j N")

delta = Real("delta")

A, B, c, d = Reals("A B c d")
#s.add(A == 1) # safe if A < B; buggy (multiple processes in critical section) if A >= B
#s.add(B == 10)
s.add(And(A > 0, B > 0))

# PROTOCOL IS BUGGY IF A >= B: try changing to A < B
s.add(A > B) # you don't have to pick actual values for A or B

s.add(c == 1)
s.add(d == 2)


Nv = 2 # number of processes

s.add(N == Nv)

reachedList = []
reachedAllList = []

#pre-allocate all variables (BAD)
#for k in range(length):
#	x.append( Function('x' + str(k), IntSort(), RealSort()) ) # continuous variable
#	q.append( Function('q' + str(k), IntSort(), ControlLocation) ) # control location
#	g.append( Int("g" + str(k)) ) # global variable
#	
#	s.add(ForAll([i], Implies(i > N, q[k](i) == ControlLocation.idle) )) # set all other indices to be idle
#	
#	s.add(ForAll([i], x[k](i) >= 0)) # all processes have nonnegative clocks at all steps
#	s.add(And( g[k] >= 0, g[k] <= N) ) # global variable is either a process id 1, ... N, or none of them (0)
#	reachedList.append( [] ) # append empty list (stores reach sets this direction)

# allocate 1st vars
for k in range(2):
	x.append( Function('x' + str(k), IntSort(), RealSort()) ) # continuous variable
	q.append( Function('q' + str(k), IntSort(), ControlLocation) ) # control location
	g.append( Int("g" + str(k)) ) # global variable

	s.add(ForAll([i], Implies(i > N, q[k](i) == ControlLocation.idle) )) # set all other indices to be idle

	s.add(ForAll([i], x[k](i) >= 0)) # all processes have nonnegative clocks at all steps
	s.add(And( g[k] >= 0, g[k] <= N) ) # global variable is either a process id 1, ... N, or none of them (0)
	reachedList.append( [] ) # append empty list (stores reach sets this direction)


# bounded model checking for bound iterations
def bmc(bound):
	initList = []
	initList.append( g[0] == 0 ) # global starts as 0, no process id
	# generate initial condition over all processes
	for m in range(1,Nv+1):
		initList.append( And( q[0](m) == ControlLocation.idle, x[0](m) == 0 ) )
	if len(initList) > 1:
		init = And( initList )
	else:
		init = initList[0]
	# quantified version
	#init = ForAll([i], Implies(And(i >= 1, i <= N), And( q[0](i) == ControlLocation.idle, x[0](i) == 0, g[0] == 0 ) ) )
	
	reachedOld = False # no states reached
	reachedOldAll = init
	reachedAll = init
	#reachedAllList.append( reachedAll )
	reached = init
	reachedList[0].append(init)
	
	terminate = 0
	
	# k is the iteration
	for k in range(length):
	
		# allocate new variables lazily / on-the-fly
		km = k + 2
		print "km: " + str(km)
		x.append( Function('x' + str(km), IntSort(), RealSort()) ) # continuous variable
		print x
		q.append( Function('q' + str(km), IntSort(), ControlLocation) ) # control location
		g.append( Int("g" + str(km)) ) # global variable
		s.add(ForAll([i], Implies(i > N, q[km](i) == ControlLocation.idle) )) # set all other indices to be idle
		s.add(ForAll([i], x[km](i) >= 0)) # all processes have nonnegative clocks at all steps
		s.add(And( g[km] >= 0, g[km] <= N) ) # global variable is either a process id 1, ... N, or none of them (0)
		reachedList.append( [] ) # append empty list (stores reach sets this direction)

		if terminate >= 1 or len(reachedList[k]) == 0:
			print "Termination condition reached"
			break
		
		# write bad states in terms of the k^th iteration states
		badList = []
		if Nv > 1:
			for m in range(1,Nv+1):
				for n in range(m,Nv+1):
					if n != m:
						badList.append( And( q[k](m) == ControlLocation.cs, q[k](n) == ControlLocation.cs ) )
		else:
			badList.append( q[k](1) == ControlLocation.cs )
		if len(badList) > 1:
			bad = Or( badList )
		else:
			bad = badList[0]
		# quantified version
		#bad = Exists([i,j], And( i >= 1, i <= N, j >= 1, j <= N, i != j, q[k](i) == ControlLocation.cs, q[k](j) == ControlLocation.cs ) )
		
		nt = len(reachedList[k])
		rt = 1
		
		reachedOldAll = reachedAll
		if len(reachedList[k]) > 1:
			reachedAll = Or(reachedList[k])
		else:
			reachedAll = reachedList[k][0]
		reachedAllList.append( reachedAll )
		
		
		if k >= bound - 1:
			print "Terminated without finding a path to bad states after " + str( k ) + " iterations."
			break

		reachedBad = And(reachedAll, bad)
		if opt_debug:
			print reachedBad
			print "\n\n\n"
		s.push() # save context
		s.add(reachedBad)
		result = s.check()
		#print result
		s.pop() # restore context

		if result == sat:
			print "UNSAFE"
			print "Old:\n"
			print reachedOldAll
			print "\nNew:\n"
			print reachedAll
			print "\nBad:\n"
			print reachedBad
			print s.model()
			print "Bad states reached after " + str( k ) + " iterations."
			print bad
			break
		
		if k >= 1:
			s.push() # save context
			if opt_debug:
				print "old:\n\n"
				print reachedOldAll
				print "new:\n\n"
				print reachedAll
				print "\n\n\n"
			#fixedPoint = Implies( reachedOldAll, reachedAll )
			#fixedPoint = Implies( reachedAllList[k-1], reachedAllList[k] ) # TODO: this should be the other way, k => k-1, but there's a problem: we need to rename states
			fixedPoint = Implies( reachedAllList[k], substitute(reachedAllList[k-1], (q[k-1](1),q[k](1)), (q[k-1](2),q[k](2)), (x[k-1](1),x[k](1)), (x[k-1](2),x[k](2)), (g[k-1],g[k]) ) ) 
			
			
			s.add( Not(fixedPoint) )
			result = s.check()
			s.pop() # restore context
		
		for reached in reachedList[k]:
			print "k=" + str(k) + " nt=" + str(nt) + " rt=" + str(rt)
			if opt_debug:
				print reached
			
			if k >= 1 and result == unsat:# and rt >= nt:
				print "Old:\n"
				print reachedOldAll
				print "\nNew:\n"
				print reachedAll
				print "Fixedpoint:\n"
				print fixedPoint
				terminate = 1
				print "SAFE after " + str(k) + " iterations"
				break
			else:
				reachedOld = reached
				ts = []
				if k < bound - 1: # termination
					
					goal = Goal()
					tt = timeTransition(k)
					
					# eliminate quantifiers (time elapse variable delta)
					goal.add( tt )
					tac = Tactic('qe')
					r = tac(goal)
					
					if opt_debug:
						print r
					
					ts.append( simplify(r.as_expr()) )
					
					#ts.append( tt ) # without q.e.

					for iv in range(1,Nv+1):
						ts.extend( stepTransition(k,iv) )
				#tnum = 0
				for t in ts:
					treached = simplify(And(reached, t))
					s.push()
					s.add(treached)
					res = s.check()
					if opt_debug:
						print "transition can be taken? " + str(res)
					s.pop()
					
					#s.push()
					#s.add( Not( Implies( treached, reachedAllList[k] ) ) )
					#resfp = s.check()
					#if resfp == unsat:
					#	print "no new states " + str(resfp)
					#else:
					#	print "NEW STATES"
					#s.pop()
					
					# only explore enabled transitions, no reason to continue down a branch if transitions cannot be taken
					if res == sat:
						reachedList[k+1].append( treached )
						#print "transition " + str(tnum) + " can be taken? " + str(res)
					#tnum = tnum + 1
			rt = rt + 1

	return 0

# transition relation for iteration k of process i 
def stepTransition(k, i):
	ts = []
	
	# idle -> cs (buggy example)
	#ts.append(And( q[k](i) == ControlLocation.idle, q[k+1](i) == ControlLocation.cs, identityTransition(k,i,[q,x],[g]) ) )
	#ts.append(And( q[k](i) == ControlLocation.idle, q[k+1](i) == ControlLocation.cs, ForAll([j], Implies( And(j >= 1, j <= N, Distinct(i,j)), And( q[k](j) == q[k+1](j), x[k](j) == x[k+1](j), g[k] == g[k+1] ) ) ) ) )
	
	# idle -> start	
	ts.append(And( q[k](i) == ControlLocation.idle, g[k] == 0, x[k+1](i) == 0, q[k+1](i) == ControlLocation.start, identityTransition(k,i,[q,x],[g])  ) )
	#ts.append(And( q[k](i) == ControlLocation.idle, g[k] == 0, x[k+1](i) == 0, q[k+1](i) == ControlLocation.start, ForAll([j], Implies( And(j >= 1, j <= N, Distinct(i,j)), And( q[k](j) == q[k+1](j), x[k](j) == x[k+1](j), g[k] == g[k+1] ) ) ) ) )

	# start -> check
	ts.append(And( q[k](i) == ControlLocation.start, x[k](i) <= A, x[k+1](i) == 0, q[k+1](i) == ControlLocation.check, g[k+1] == i, identityTransition(k,i,[q,x],[])  ) )
	#ts.append(And( q[k](i) == ControlLocation.start, x[k](i) <= A, x[k+1](i) == 0, q[k+1](i) == ControlLocation.check, g[k+1] == i,          ForAll([j], Implies( And(Distinct(i,j), j >= 1, j <= N), And(q[k](j) == q[k+1](j), x[k](j) == x[k+1](j) )))                ) )
	
	# check -> cs
	ts.append(And( q[k](i) == ControlLocation.check, g[k] == i, x[k](i) >= B, q[k+1](i) == ControlLocation.cs, identityTransition(k,i,[q,x],[g])          ) )
	#ts.append(And( q[k](i) == ControlLocation.check, g[k] == i, x[k](i) >= B, q[k+1](i) == ControlLocation.cs,      ForAll([j], Implies( And(j >= 1, j <= N, Distinct(i,j)), And(q[k](j) == q[k+1](j), x[k](j) == x[k+1](j), g[k] == g[k+1] )))           ) )
	
	# check -> idle
	ts.append(And( q[k](i) == ControlLocation.check, g[k] != i, x[k](i) >= B, q[k+1](i) == ControlLocation.idle, identityTransition(k,i,[q,x],[g])      ) )
	#ts.append(And( q[k](i) == ControlLocation.check, g[k] != i, x[k](i) >= B, q[k+1](i) == ControlLocation.idle,        ForAll([j], Implies( And(j >= 1, j <= N, Distinct(i,j)), And( q[k](j) == q[k+1](j), x[k](j) == x[k+1](j), g[k] == g[k+1] )))               ) )
	
	# cs -> idle
	ts.append(And( q[k](i) == ControlLocation.cs, q[k+1](i) == ControlLocation.idle, g[k+1] == 0,  identityTransition(k,i,[q,x],[])        ) )
	#ts.append(And( q[k](i) == ControlLocation.cs, q[k+1](i) == ControlLocation.idle, g[k+1] == 0,        ForAll([j], Implies( And(j >= 1, j <= N, Distinct(i,j) ), And( q[k](j) == q[k+1](j), x[k](j) == x[k+1](j) )))         ) )
	
	return ts

def timeTransition(k):
	# quantified indices version
	#return Exists([delta], And(delta > 0, ForAll([j], And( q[k+1](j) == q[k](j), g[k+1] == g[k], x[k+1](j) == x[k](j) + delta )) ) ) # doesn't respect invariants
	
	ts = []
	for j in range(1,Nv+1):
		# all discrete variables remain unchanged; the last implication enforces the invariant in the start location
		ts.append( And( q[k+1](j) == q[k](j), g[k+1] == g[k], x[k+1](j) == x[k](j) + delta, Implies( (q[k+1](j) == ControlLocation.start), x[k+1](j) <= A ) ) )
	
	p = Exists([delta], And(delta > 0, And( ts ) ) )
	
	return p

# state identity between states in execution
# i.e., if process 1 makes a discrete transition, assert that the states of process 2 do not change, e.g., q_2 at state k+1 = q_2 at state k
def identityTransition(k,i,lvars,gvars):
	idList = []
	for m in range(1,Nv+1):
		if i != m:
			#p = And( q[k](m) == q[k+1](m), x[k](m) == x[k+1](m), g[k] == g[k+1] )
			pl = []
			for lv in lvars:
				pl.append( lv[k](m) == lv[k+1](m) )
			for gv in gvars:
				pl.append( gv[k] == gv[k+1] )
			if len(pl) > 1:
				p = And( pl )
			else:
				p = pl[0]
			idList.append( p )
	
	if len(idList) == 0:
		idAll = True
	elif len(idList) > 1:
		idAll = And( idList )
	else:
		idAll = idList[0]
	return idAll



# call BMC for length iterations
bmc(length)


